/*
 * Copyright (c) 2021 DuckDuckGo
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.duckduckgo.app.email.waitlist

import android.util.Log
import androidx.lifecycle.LifecycleOwner
import androidx.test.platform.app.InstrumentationRegistry
import androidx.work.Configuration
import androidx.work.OneTimeWorkRequestBuilder
import androidx.work.WorkInfo
import androidx.work.WorkManager
import androidx.work.impl.utils.SynchronousExecutor
import androidx.work.testing.WorkManagerTestInitHelper
import com.duckduckgo.app.CoroutineTestRule
import com.duckduckgo.app.email.EmailManager
import com.duckduckgo.app.job.TestWorker
import com.duckduckgo.app.notification.NotificationSender
import com.duckduckgo.app.notification.model.SchedulableNotification
import com.duckduckgo.app.waitlist.email.AppEmailWaitlistCodeFetcher
import com.duckduckgo.app.waitlist.email.EmailWaitlistCodeFetcher
import com.duckduckgo.app.waitlist.email.EmailWaitlistWorkRequestBuilder
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.test.TestScope
import kotlinx.coroutines.test.runTest
import org.junit.Assert.assertFalse
import org.junit.Assert.assertTrue
import org.junit.Before
import org.junit.Rule
import org.junit.Test
import org.mockito.kotlin.mock
import org.mockito.kotlin.never
import org.mockito.kotlin.verify
import org.mockito.kotlin.whenever
import java.util.concurrent.TimeUnit

@ExperimentalCoroutinesApi
class AppEmailWaitlistCodeFetcherTest {

    @get:Rule
    var coroutineRule = CoroutineTestRule()

    private val context = InstrumentationRegistry.getInstrumentation().targetContext
    private val mockEmailManager: EmailManager = mock()
    private val mockNotification: SchedulableNotification = mock()
    private val mockNotificationSender: NotificationSender = mock()
    private val mockOwner: LifecycleOwner = mock()

    private lateinit var workManager: WorkManager
    private lateinit var testee: EmailWaitlistCodeFetcher

    @Before
    fun before() {
        initializeWorkManager()
        testee = AppEmailWaitlistCodeFetcher(
            workManager,
            mockEmailManager,
            mockNotification,
            mockNotificationSender,
            coroutineRule.testDispatcherProvider,
            TestScope()
        )
    }

    @Test
    fun whenFetchInviteCodeIfUserInQueueCodeAlreadyExistsAndWorkerIsRunningThenCancelWorker() = runTest {
        givenUserIsInTheQueueAndCodeAlreadyExists()
        enqueueWaitlistWorker()

        testee.fetchInviteCode()

        assertWaitlistWorkerIsNotEnqueued()
    }

    @Test
    fun whenFetchInviteCodeIfUserInQueueAndCodeReturnedThenCancelWorker() = runTest {
        givenUserIsInTheQueueAndCodeReturned()
        enqueueWaitlistWorker()

        testee.fetchInviteCode()

        assertWaitlistWorkerIsNotEnqueued()
    }

    @Test
    fun whenFetchInviteCodeIfUserInQueueAndCodeReturnedThenNotificationSent() = runTest {
        givenUserIsInTheQueueAndCodeReturned()
        enqueueWaitlistWorker()

        testee.fetchInviteCode()

        verify(mockNotificationSender).sendNotification(mockNotification)
    }

    @Test
    fun whenFetchInviteCodeIfUserInQueueAndNoCodeReturnedThenDoNothing() = runTest {
        givenUserIsInTheQueueAndNoCodeReturned()
        enqueueWaitlistWorker()

        testee.fetchInviteCode()

        verify(mockNotificationSender, never()).sendNotification(mockNotification)
        assertWaitlistWorkerIsEnqueued()
    }

    @Test
    fun whenExecuteWaitlistCodeFetcherIfUserInNotInQueueThenDoNothing() = runTest {
        whenever(mockEmailManager.waitlistState()).thenReturn(EmailManager.WaitlistState.NotJoinedQueue)

        (testee as AppEmailWaitlistCodeFetcher).onStart(mockOwner)

        verify(mockEmailManager, never()).fetchInviteCode()
    }

    @Test
    fun whenExecuteWaitlistCodeFetcherIfUserIsInBetaThenDoNothing() = runTest {
        whenever(mockEmailManager.waitlistState()).thenReturn(EmailManager.WaitlistState.InBeta)

        (testee as AppEmailWaitlistCodeFetcher).onStart(mockOwner)

        verify(mockEmailManager, never()).fetchInviteCode()
    }

    private fun givenUserIsInTheQueueAndCodeAlreadyExists() = runTest {
        whenever(mockEmailManager.waitlistState()).thenReturn(EmailManager.WaitlistState.JoinedQueue())
        whenever(mockEmailManager.fetchInviteCode()).thenReturn(EmailManager.FetchCodeResult.CodeExisted)
    }

    private fun givenUserIsInTheQueueAndCodeReturned() = runTest {
        whenever(mockEmailManager.waitlistState()).thenReturn(EmailManager.WaitlistState.JoinedQueue())
        whenever(mockEmailManager.fetchInviteCode()).thenReturn(EmailManager.FetchCodeResult.Code)
    }

    private fun givenUserIsInTheQueueAndNoCodeReturned() = runTest {
        whenever(mockEmailManager.waitlistState()).thenReturn(EmailManager.WaitlistState.JoinedQueue())
        whenever(mockEmailManager.fetchInviteCode()).thenReturn(EmailManager.FetchCodeResult.NoCode)
    }

    private fun initializeWorkManager() {
        val config = Configuration.Builder()
            .setMinimumLoggingLevel(Log.DEBUG)
            .setExecutor(SynchronousExecutor())
            .build()

        WorkManagerTestInitHelper.initializeTestWorkManager(context, config)
        workManager = WorkManager.getInstance(context)
    }

    private fun enqueueWaitlistWorker() {
        val requestBuilder = OneTimeWorkRequestBuilder<TestWorker>()
        val request = requestBuilder
            .addTag(EmailWaitlistWorkRequestBuilder.EMAIL_WAITLIST_SYNC_WORK_TAG)
            .setInitialDelay(10, TimeUnit.SECONDS)
            .build()
        workManager.enqueue(request)
    }

    private fun assertWaitlistWorkerIsEnqueued() {
        val scheduledWorkers = getScheduledWorkers()
        assertFalse(scheduledWorkers.isEmpty())
    }

    private fun assertWaitlistWorkerIsNotEnqueued() {
        val scheduledWorkers = getScheduledWorkers()
        assertTrue(scheduledWorkers.isEmpty())
    }

    private fun getScheduledWorkers(): List<WorkInfo> {
        return workManager
            .getWorkInfosByTag(EmailWaitlistWorkRequestBuilder.EMAIL_WAITLIST_SYNC_WORK_TAG)
            .get()
            .filter { it.state == WorkInfo.State.ENQUEUED }
    }
}
